# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import warnings

from dbnd._core.configuration.environ_config import (
    get_dbnd_project_config,
    is_unit_test_mode,
    should_fix_pyspark_imports,
)
from dbnd._core.utils.basics.signal_utils import register_graceful_sigterm
from dbnd._core.utils.platform.osx_compatible.requests_in_forked_process import (
    enable_osx_forked_request_calls,
)
from dbnd_run.airflow.airflow_env import _initialize_airflow_home


def _surpress_loggers():
    logging.getLogger("googleapiclient.discovery_cache").setLevel(logging.ERROR)
    logging.getLogger("googleapiclient").setLevel(logging.WARN)


def _suppress_warnings():
    warnings.simplefilter("ignore", FutureWarning)


def fix_pyspark_imports():
    import sys

    pyspark_libs, regular_libs = [], []
    for p in sys.path:
        if "spark-core" in p or "pyspark.zip" in p:
            pyspark_libs.append(p)
        else:
            regular_libs.append(p)
    regular_libs.extend(pyspark_libs)
    sys.path = regular_libs


def dbnd_disable_databand_dags_loading(dbnd_config):
    """
    if we are running from "dbnd" entrypoint, we probably do not need to load Scheduled DAG
    this will prevent from every airflow command to access dbnd web api
    """

    if dbnd_config.getboolean("airflow", "auto_disable_scheduled_dags_load"):
        os.environ["DBND_DISABLE_SCHEDULED_DAGS_LOAD"] = "True"


def _dbnd_bootstrap_plugins():
    project_config = get_dbnd_project_config()

    if project_config.is_no_plugins:
        return

    from dbnd import dbnd_config
    from dbnd_run.plugin.dbnd_plugins import (
        pm,
        register_dbnd_plugins,
        register_dbnd_user_plugins,
    )

    if not project_config.disable_pluggy_entrypoint_loading:
        register_dbnd_plugins()

    user_plugins = dbnd_config.get("core", "plugins", None)
    if user_plugins:
        register_dbnd_user_plugins(user_plugins.split(","))

    if is_unit_test_mode():
        pm.hook.dbnd_setup_unittest()

    pm.hook.dbnd_setup_plugin()


_dbnd_bootstrap_orchestration_status = None


def dbnd_bootstrap_orchestration():
    """Runs dbnd bootstrapping."""
    global _dbnd_bootstrap_orchestration_status
    if _dbnd_bootstrap_orchestration_status is not None:
        return
    try:
        _dbnd_bootstrap_orchestration_status = "loading"
        from dbnd import dbnd_config

        _initialize_airflow_home()

        _dbnd_bootstrap_plugins()

        # now we can run user code ( at driver/task)
        from dbnd._core.configuration import environ_config
        from dbnd._core.utils.basics.load_python_module import run_user_func

        user_preinit = environ_config.get_user_preinit()
        if user_preinit:
            run_user_func(user_preinit)

        _surpress_loggers()
        _suppress_warnings()

        register_graceful_sigterm()

        if should_fix_pyspark_imports():
            fix_pyspark_imports()

        if dbnd_config.getboolean("run", "fix_env_on_osx"):
            enable_osx_forked_request_calls()

        _dbnd_bootstrap_orchestration_status = "loaded"
    except Exception as ex:
        _dbnd_bootstrap_orchestration_status = "error: %s" % str(ex)
        raise ex



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.task.config import Config
from dbnd_run.task.data_source_task import DataSourceTask, data_combine, data_source
from dbnd_run.task.pipeline_task import PipelineTask
from dbnd_run.task.python_task import PythonTask
from dbnd_run.task.task import Task
from dbnd_run.tasks import basics


__all__ = [
    "Task",
    "Config",
    "data_combine",
    "data_source",
    "DataSourceTask",
    "PipelineTask",
    "PythonTask",
    "basics",
]



# © Copyright Databand.ai, an IBM Company 2022

import typing

from typing import Union

from dbnd._core.errors import DatabandConfigError
from dbnd._core.errors.base import ConfigLookupError
from dbnd._core.settings.core import DatabandSystemConfig
from dbnd._core.settings.histogram import HistogramConfig  # noqa: F401
from dbnd._core.settings.run_info import RunInfoConfig  # noqa: F401
from dbnd._core.task import Config
from dbnd._core.task_build.task_registry import build_task_from_config
from dbnd_run.run_settings.describe import DescribeConfig
from dbnd_run.run_settings.engine import EngineConfig  # noqa: F401
from dbnd_run.run_settings.env import EnvConfig, LocalEnvConfig  # noqa: F401
from dbnd_run.run_settings.git import GitConfig
from dbnd_run.run_settings.log import RunLoggingConfig
from dbnd_run.run_settings.output import OutputConfig
from dbnd_run.run_settings.run import RunConfig
from dbnd_run.run_settings.scheduler import SchedulerConfig


if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext


class RunSettings(object):
    def __init__(self, databand_context):
        super(RunSettings, self).__init__()
        self.databand_context = databand_context  # type: DatabandContext

        from dbnd._core.tracking.registry import register_store
        from dbnd_run.orchestration_tracking.backends.tracking_store_file import (
            FileTrackingStore,
        )

        register_store("file", FileTrackingStore)

        self.run = RunConfig()

        self.env = self.get_env_config(self.system.env)
        databand_context.config.set_values(
            config_values={"task": {"task_env": self.system.env}},
            source="RunSettings.env",
        )

        self.git = GitConfig()

        self.describe = DescribeConfig()

        self.output = OutputConfig()
        self.run_logging = RunLoggingConfig()
        self.scheduler = SchedulerConfig()

        self.singleton_configs = {}

        self.user_configs = {}
        for user_config in self.run.user_configs:
            self.user_configs[user_config] = build_task_from_config(user_config)

    @property
    def system(self):
        # type:()->DatabandSystemConfig
        return self.databand_context.system_settings

    def get_env_config(self, name_or_env):
        # type: ( Union[str, EnvConfig]) -> EnvConfig
        if isinstance(name_or_env, EnvConfig):
            return name_or_env

        enabled_environments = self.run.environments
        if name_or_env not in enabled_environments:
            raise DatabandConfigError(
                "Unknown env name '%s', available environments are %s,  please enable it at '[core]environments' "
                % (name_or_env, enabled_environments)
            )
        return build_task_from_config(name_or_env, EnvConfig)

    def get_config(self, name):
        """
        Retrieve a config by name, return existing instance or new one if not created yet.
        The config lookup is using **only the directed subclass of Config object**.
        """
        if hasattr(self, name):
            # dbnd-core configs are attributes of settings.
            # we don't want 2 instances of each config per context.
            config = getattr(self, name)
            if isinstance(config, Config):
                return config

        if name in self.singleton_configs:
            return self.singleton_configs[name]

        # works only for a direct child of Config
        for conf_cls in Config.__subclasses__():
            if conf_cls._conf__task_family == name:
                conf = conf_cls()
                self.singleton_configs[name] = conf
                return conf

        raise ConfigLookupError("Couldn't found config - {}".format(name))



# © Copyright Databand.ai, an IBM Company 2022
from __future__ import absolute_import

import logging
import os

from dbnd._core.log import dbnd_log_debug
from dbnd._core.utils.object_utils import patch_module_attr
from dbnd_run.airflow.compat import AIRFLOW_VERSION_2
from dbnd_run.errors.versioned_dagbag import failed_to_load_versioned_dagbag_plugin


logger = logging.getLogger(__name__)


# WE SHOULD NOT HAVE ANY dbnd imports here -- circle import:  dbnd_airflow -> airflow -> load .. -> import dbnd_airflow


def _use_databand_airflow_dagbag(app, views):
    """
    Overriding Airflow Dagbag, so versioned dags can be used
    :return:
    """
    import airflow

    from airflow import settings

    from dbnd_run.airflow.web.databand_versioned_dagbag import (
        DbndAirflowDagBag,
        DbndDagModel,
    )

    if os.environ.get("SKIP_DAGS_PARSING") != "True":
        views.dagbag = DbndAirflowDagBag(settings.DAGS_FOLDER)
    else:
        views.dagbag = DbndAirflowDagBag(os.devnull, include_examples=False)

    # some views takes dag from dag model
    if hasattr(views, "DagModel"):
        patch_module_attr(views, "DagModel", DbndDagModel)

    # dag_details invoke DagBag directly
    patch_module_attr(airflow.models, "DagBag", DbndAirflowDagBag)
    patch_module_attr(airflow.models.dag, "DagBag", DbndAirflowDagBag)
    patch_module_attr(airflow.models.dag, "DagModel", DbndDagModel)

    logging.info(
        "Using DBND DagBag with support for versioned dags and historic dag runs."
    )

    if hasattr(app, "dag_bag"):
        if os.environ.get("SKIP_DAGS_PARSING") == "True":
            app.dag_bag = DbndAirflowDagBag(os.devnull, include_examples=False)
        else:
            from airflow.settings import DAGS_FOLDER

            app.dag_bag = DbndAirflowDagBag(DAGS_FOLDER, read_dags_from_db=True)


def patch_airflow_create_app():
    dbnd_log_debug("Adding support for versioned DBND DagBag")

    def patch_create_app_in_module(web_module):
        original_create_app_func = web_module.create_app
        if hasattr(web_module, "_patched_versioned_dag"):
            return web_module

        def patched_create_app(*args, **kwargs):
            res = original_create_app_func(*args, **kwargs)
            if AIRFLOW_VERSION_2:
                from airflow.www import views

                app = res
            else:
                from airflow import settings

                if settings.RBAC:
                    from airflow.www_rbac import views
                else:
                    from airflow.www import views
                app = res[0]

            try:
                _use_databand_airflow_dagbag(app, views)
            except Exception as e:
                raise failed_to_load_versioned_dagbag_plugin(e)

            return res

        web_module.create_app = patched_create_app
        setattr(web_module, "_patched_versioned_dag", True)
        return web_module

    if AIRFLOW_VERSION_2:
        from airflow.www import app as airflow_app

        patch_create_app_in_module(airflow_app)
    else:
        from airflow.www_rbac import app as airflow_app_rbac

        patch_create_app_in_module(airflow_app_rbac)

        from airflow.www import app as airflow_app_no_rbac

        patch_create_app_in_module(airflow_app_no_rbac)


def create_app(config=None, testing=False):
    patch_airflow_create_app()

    if AIRFLOW_VERSION_2:
        from airflow.www import app as airflow_app

        app = airflow_app.create_app(config=config, testing=testing)
        appbuilder = app.appbuilder
    else:
        from airflow.www_rbac import app as airflow_app

        app, appbuilder = airflow_app.create_app(config=config, testing=testing)

    logging.info("Airflow applications has been created")
    return app, appbuilder



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors import DatabandConfigError


def parallel_or_remote_sqlite(executor):  # type: (str) -> DatabandConfigError
    return DatabandConfigError(
        "'%s' executor is not supported when using an sqlite database" % executor,
        help_msg="Please switch to a different database for all parallel and "
        "remote executions (we recommend PostgreSQL)",
    )


def parallel_with_inprocess(executor_type):
    return DatabandConfigError(
        "We don not support parallel execution "
        "in {executor_type} executor type".format(executor_type=executor_type)
    )



# © Copyright Databand.ai, an IBM Company 2022

import hashlib
import logging
import os

from dbnd._core.constants import _TaskDbndRunDataSource
from dbnd._core.errors import friendly_error
from dbnd._core.parameter.parameter_builder import output
from dbnd._core.utils.task_utils import to_targets
from dbnd._core.utils.traversing import flatten
from dbnd_run.task.task import Task
from targets.multi_target import MultiTarget


logger = logging.getLogger(__name__)


class DataSourceTask(Task, _TaskDbndRunDataSource):
    """Subclass for references to data sources."""

    _conf__task_type_name = "data"
    task_target_date = None


class _DataSource(DataSourceTask):
    data = output

    def run(self):

        missing = []
        for partial_output in flatten(self.data):
            if not partial_output.exists():
                missing.append(partial_output)
        if missing:
            raise friendly_error.task_data_source_not_exists(
                self, missing, downstream=self.task_dag.downstream
            )


def data_source(data):
    a, b = os.path.split(str(data))
    a = hashlib.md5(a.encode("utf-8")).hexdigest()[:6]  # nosec B324
    logger.debug("Creating external task %s", data)
    return _DataSource(data=data, task_name="file_%s__%s" % (a, b))


class DataTask(Task):
    pass


def data_combine(inputs, sort=False):
    targets = flatten(to_targets(inputs))
    if sort:
        targets = sorted(targets, key=lambda x: x.path)
    data = MultiTarget(targets)
    return data



# © Copyright Databand.ai, an IBM Company 2022

from dbnd.errors import (
    DatabandBuildError,
    DatabandConfigError,
    DatabandError,
    DatabandRunError,
    DatabandRuntimeError,
    DatabandSystemError,
)


__all__ = [
    "DatabandBuildError",
    "DatabandConfigError",
    "DatabandError",
    "DatabandRunError",
    "DatabandRuntimeError",
    "DatabandSystemError",
]



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import re

from targets.utils.path import no_trailing_slash


logger = logging.getLogger(__name__)
windows_drive_re = re.compile(r"^[A-Z]\:\\")


def calculate_path(task, name, path_pattern, output_ext="", is_dir=False):
    # do we need to have a path like date=None ?
    # path_vars = [str(v) if v is not None else "" for v in task.get_vars()]
    path_vars = task.get_template_vars()

    sep = "/"

    root = no_trailing_slash(str(task.get_root()))
    if windows_drive_re.match(root):
        sep = os.sep
    task_env = task.task_env
    path = path_pattern.format(
        root=root,
        dbnd_root=str(task_env.dbnd_root),
        output_name=name,
        output_ext=output_ext or "",
        sep=sep,
        env_label=task_env.env_label,
        **path_vars
    )
    if is_dir:
        path += "/"
    return path



# © Copyright Databand.ai, an IBM Company 2022


def run_task(task):
    run = task.dbnd_run()
    from IPython.core.display import HTML

    return HTML(
        'You can review Databand Run <a href="%s" target="_blank" >at this link</a>'
        % run.run_url
    )



# © Copyright Databand.ai, an IBM Company 2022

from collections import Counter, defaultdict
from typing import List

from airflow.models import TaskInstance
from airflow.utils.db import provide_session


class AirflowTaskInstanceStateManager(object):
    """
    AirflowTaskInstanceStateManager holds latest state info for all relevant task_instances
    """

    def __init__(self):
        self.status = defaultdict(dict)

    def _get_dag_run(self, dag_id, execution_date):
        return self.status[(dag_id, execution_date)]

    def refresh_from_db(self, dag_id, execution_date, session):
        TI = TaskInstance
        updated_status = (
            session.query(TI.task_id, TI.state)
            .filter(TI.dag_id == dag_id, TI.execution_date == execution_date)
            .all()
        )

        self.status[(dag_id, execution_date)] = dict(updated_status)

    def get_state(self, dag_id, execution_date, task_id):
        return self._get_dag_run(dag_id, execution_date).get(task_id)

    def get_aggregated_state_status(self, dag_id, execution_date, task_ids):
        status = self._get_dag_run(dag_id, execution_date)
        return Counter(status.get(task_id) for task_id in task_ids)

    def sync_to_object(self, task_instances):
        # type: (List[TaskInstance]) ->None

        for ti in task_instances:
            ti.state = self.get_state(ti.dag_id, ti.execution_date, ti.task_id)

    @provide_session
    def refresh_task_instances_state(
        self, task_instances, dag_id, execution_date, session=None
    ):
        self.refresh_from_db(dag_id, execution_date, session)
        self.sync_to_object(task_instances)



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging
import random
import time

from dbnd._core.task_build.dbnd_decorator import pipeline, task
from dbnd._core.tracking.metrics import log_metric


logger = logging.getLogger(__name__)


@task
def dbnd_sanity_check(check_time=datetime.datetime.now(), fail_chance=0.0):
    # type: (datetime.datetime, float) -> str
    logger.info("Running Databand Task Sanity Check!")

    log_metric("metric_check", "OK")

    if random.random() < fail_chance:
        raise Exception("Failed accordingly to fail_chance: %f" % fail_chance)

    log_metric("Happiness Level", "High")

    logger.info("Your system is good to go! Enjoy Databand!")
    return "Databand has been checked at %s" % check_time


@task
def dbnd_simple_task(
    check_time=datetime.datetime.now(), fail_chance=0.0, sleep=0.0, loop=1
):  # type: (datetime.datetime, float, float, int) -> str
    logger.info("Running dbnd_simple_task!")
    log_metric("metric_check", "OK")
    log_metric("metric_random_value", random.random())

    for current in range(loop):
        if current:
            logger.info("Loop %s out of %s", current, loop)
        if sleep:
            logger.info("Sleeping for %s seconds", sleep)
            time.sleep(sleep)

        if random.random() < fail_chance:
            raise Exception("Failed accordingly to fail_chance: %f" % fail_chance)

    return "Databand has been checked at %s" % check_time


@pipeline
def dbnd_simple_parallel_pipeline(tasks_num=2):
    return {
        str(i): dbnd_simple_task(task_name=f"dbnd_simple_task_{i}")
        for i in range(tasks_num)
    }


@task
def dbnd_sub_run():
    logger.info("STARTING THE RUN")
    from dbnd import get_databand_context

    databand_context = get_databand_context()
    databand_context.run_settings.run.submit_driver = False
    databand_context.run_settings.run.is_archived = True
    from dbnd._core.utils.timezone import utcnow

    databand_context.run_settings.run.execution_date = utcnow()

    result = dbnd_simple_task.dbnd_run(
        task_name="dbnd_sub_simple_run", task_version=str(utcnow())
    )

    logger.info("FINISHED THE RUN")

    return result.root_task.result.load(str)


@pipeline
def dbnd_pipeline_with_sub(tasks_num=2):
    tasks = {}
    # tasks = {
    #     str(i): dbnd_simple_task(task_name=f"dbnd_simple_task_{i}")
    #     for i in range(tasks_num)
    # }
    tasks["sub_run"] = dbnd_sub_run()
    return tasks



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import logging
import os

from distutils.dir_util import copy_tree

from dbnd._core.configuration.environ_config import get_dbnd_project_config
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._vendor import click
from dbnd_run.utils.dbnd_run_module import get_dbnd_run_conf_file


logger = logging.getLogger(__name__)


@click.command()
@click.option("--overwrite", is_flag=True)
@click.option("--dbnd-home", envvar="DBND_HOME", show_envvar=True)
@click.option("--dbnd-system", envvar="DBND_SYSTEM", show_envvar=True)
@click.pass_context
def project_init(ctx, overwrite, dbnd_home, dbnd_system):
    """Initialize the project structure"""
    dbnd_bootstrap(enable_dbnd_run=True)

    from dbnd._core.errors import DatabandSystemError

    if not dbnd_home:
        dbnd_home = get_dbnd_project_config().dbnd_home()

    os.environ["SKIP_DAGS_PARSING"] = "True"  # Exclude airflow dag examples

    conf_folder = get_dbnd_run_conf_file("project_init")

    if os.path.exists(os.path.join(dbnd_home, "project.cfg")):
        if not overwrite:
            raise DatabandSystemError(
                "You are trying to re-initialize your project. You already have dbnd configuration at %s. "
                "You can force project-init by providing --overwrite flag. "
                "If you need to create/update database use `dbnd-web db init` instead"
                % dbnd_system
            )

        logger.warning(
            "You are re-initializing your project, all files at %s are going to be over written!"
            % dbnd_home
        )

    copy_tree(conf_folder, dbnd_home)
    click.echo("Databand project has been initialized at %s" % dbnd_home)
    return



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

import six

from dbnd._core.constants import TaskEssence
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.parameter.parameter_value import ParameterFilters
from dbnd._core.task_ctrl.task_ctrl import TaskSubCtrl
from dbnd._core.utils.basics.nothing import is_defined
from dbnd._core.utils.basics.text_banner import safe_string


if typing.TYPE_CHECKING:
    pass
logger = logging.getLogger(__name__)

_MAX_PARAM_VALUE_AT_DB = 3000


def _safe_params(params):
    return [
        (name, safe_string(value, _MAX_PARAM_VALUE_AT_DB)) for name, value in params
    ]


def _parameter_value_to_argparse_str(p, p_value):
    formatted_value = p.as_str_input(p_value)
    if isinstance(p_value, list) or isinstance(p_value, dict):
        formatted_value = '"{}"'.format(formatted_value)
    if p.name == "task_env":
        return ["--env", "{}".format(formatted_value)]
    return ["--set", "{}={}".format(p.name, formatted_value)]


class TaskCallRepresentation(TaskSubCtrl):
    def __init__(self, task):
        super(TaskCallRepresentation, self).__init__(task)
        self.task_command_line = None
        self.task_functional_call = None

    def initialize(self):
        self.task_command_line = self.calculate_command_line_for_task()
        self.task_functional_call = self.calculate_task_call()

    def __get_relevant_params(self):
        relevant = []
        for p, value in self.params.get_params_with_value(ParameterFilters.USER_INPUTS):
            if is_defined(p.default):
                try:
                    same_value = value == p.default
                except Exception:
                    same_value = p.signature(value) == p.signature(p.default)
                if same_value:
                    continue
            relevant.append((p, value))
        return sorted(relevant, key=lambda x: x[0].name)

    def __get_override_repr(self):
        if not TaskEssence.ORCHESTRATION.is_instance(self.task):
            # config and overrides are exists in orchestration mode only
            return {}
        if not self.task.task_config_override:
            return {}
        result = {}
        for k, v in six.iteritems(self.task.task_config_override):
            if isinstance(k, ParameterDefinition):
                k = "%s.%s" % (k.task_definition.task_family, k.name)
            result[k] = v
        return result

    def calculate_command_line_for_task(self):

        task_name = self.task.task_family
        base = "dbnd run {task_name} ".format(task_name=task_name)

        params = []
        for p, p_value in self.__get_relevant_params():
            params.extend(_parameter_value_to_argparse_str(p, p_value))

        overrides = self.__get_override_repr()
        if overrides:
            for k, v in sorted(overrides.items()):
                overrides_str = "'{param_name}':{param_value}".format(
                    param_name=k, param_value=repr(v)
                )
                params.append('--conf "{ %s }"' % overrides_str)

        if params:
            base += " ".join(params)
        return base

    def task_repr(self):
        params = []
        for p, p_value in self.__get_relevant_params():
            param = "{param_name}={param_value}".format(
                param_name=p.name, param_value=p.to_repr(p_value)
            )
            params.append(param)

        overrides = self.__get_override_repr()
        if overrides:
            for k, v in sorted(overrides.items()):
                overrides_str = "'{param_name}':{param_value}".format(
                    param_name=k, param_value=repr(v)
                )
                params.append("overrides={ %s }" % overrides_str)

        task_ref = "{task_name}".format(
            task_name=self.task.task_definition.full_task_family
        )
        if self.task.task_decorator:
            task_ref += ".task"
        return "{task_ref}({params})".format(
            task_ref=task_ref, params=", ".join(params)
        )

    def calculate_task_call(self):
        task_repr = self.task_repr()
        return "{task_repr}.dbnd_run()".format(task_repr=task_repr)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_run.errors import execute_engine, executor_k8s, graph, task_execution


__all__ = ["execute_engine", "executor_k8s", "graph", "task_execution"]



# © Copyright Databand.ai, an IBM Company 2022

from airflow.ti_deps.deps.trigger_rule_dep import TriggerRuleDep
from airflow.utils.db import provide_session
from airflow.utils.state import State


class TriggerRuleDepOptimizied(TriggerRuleDep):
    """
    the original implementation fetching data from DB directly
    we use ti_state_manager with the latest states to calculate that
    Assumption: ti_state_manager is refreshed before we run this code.
    """

    @provide_session
    def _get_dep_statuses(self, ti, session, dep_context):
        from dbnd_run.current import try_get_run_executor

        run_executor = try_get_run_executor()

        if not run_executor or not run_executor.task_executor:
            # if we are in Scheduler or Web Server
            # we don't have current run
            # let standard not optimized implementation
            for d in super(TriggerRuleDepOptimizied, self)._get_dep_statuses(
                ti, session, dep_context
            ):
                yield d
            return

        ti_state_manager = run_executor.task_executor.get_ti_state_manager()
        from airflow.utils.trigger_rule import TriggerRule

        TR = TriggerRule

        # Checking that all upstream dependencies have succeeded
        if not ti.task.upstream_list:
            yield self._passing_status(
                reason="The task instance did not have any upstream tasks."
            )
            return

        if ti.task.trigger_rule == TR.DUMMY:
            yield self._passing_status(reason="The task had a dummy trigger rule set.")
            return

        status = ti_state_manager.get_aggregated_state_status(
            dag_id=ti.dag_id,
            execution_date=ti.execution_date,
            task_ids=ti.task.upstream_task_ids,
        )

        successes = status[State.SUCCESS]
        skipped = status[State.SKIPPED]
        failed = status[State.FAILED]
        upstream_failed = status[State.UPSTREAM_FAILED]

        for dep_status in self._evaluate_trigger_rule(
            ti=ti,
            successes=successes,
            skipped=skipped,
            failed=failed,
            upstream_failed=upstream_failed,
            done=successes + skipped + failed + upstream_failed,
            flag_upstream_failed=dep_context.flag_upstream_failed,
            session=session,
        ):
            yield dep_status



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import List

from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config


logger = logging.getLogger(__name__)


class RunConfig(config.Config):
    """Databand's per run settings (e.g. execution date)"""

    _conf__task_family = "run"

    # Executor configuration
    parallel = parameter(default=None).help("Run specific tasks in parallel.")[bool]
    task_executor_type = parameter(
        default=None,
        description="Set alternate executor type. Some of the options are `local`, `airflow_inprocess`, "
        "`airflow_multiprocess_local`, or `airflow_kubernetes`.",
    )[str]

    enable_airflow_kubernetes = parameter(
        default=True,
        description="Enable the use of Kubernetes executor for kubernetes engine submission.",
    )[bool]

    ######
    # Local/Remote control
    interactive = parameter(
        default=False,
        description="When submitting a driver to remote execution, keep track of the "
        "submitted process and wait for completion.",
    )[bool]
    submit_driver = parameter(
        description="Override `env.submit_driver` for the specific environment."
    ).none[bool]
    submit_tasks = parameter(
        description="Override `env.submit_tasks` for the specific environment."
    ).none[bool]

    # What to do on run
    open_web_tracker_in_browser = parameter(
        description="If True, opens web tracker in browser during the task's run."
    ).value(False)

    dry = parameter(default=False).help(
        "Do not execute tasks, stop before sending them to the execution, and print their status."
    )[bool]

    run_result_json_path = parameter(default=None).help(
        "Set the path to save the task band of the run."
    )[str]

    debug_pydevd_pycharm_port = parameter(default=None).help(
        "Enable debugging with `pydevd_pycharm` by setting this to the port value expecting the debugger to connect. \n"
        "This will start a new `settrace` connecting to `localhost` on the requested port, "
        "right before starting the driver task_run_executor."
    )[int]

    ######
    # AIRFLOW EXECUTOR CONFIG
    mark_success = parameter(
        description="Mark jobs as succeeded without running them."
    ).value(False)

    ######
    # Task Selectors (to schedule specific task from pipeline)
    id = parameter(default=None, description="Set the list of task IDs to run.")[
        List[str]
    ]
    selected_tasks_regex = parameter(
        default=None,
        description="Run only specified tasks. This is a regular expression.",
    )[List[str]]

    ignore_dependencies = parameter(
        description="The regex to filter specific task_ids."
    ).value(False)
    ignore_first_depends_on_past = parameter(
        description="The regex to filter specific task_ids."
    ).value(False)

    ######
    # Scheduler configuration

    skip_completed = parameter(
        description="Mark jobs as succeeded without running them."
    ).value(True)
    fail_fast = parameter(
        description="Skip all remaining tasks if a task has failed."
    ).value(True)
    enable_prod = parameter(description="Enable production tasks.").value(False)

    skip_completed_on_run = parameter(default=True).help(
        "Determine whether dbnd task should check that the task is completed and mark it as re-used on task execution."
    )[bool]

    validate_task_inputs = parameter(default=True).help(
        "Determine whether dbnd task should check that all input files exist."
    )[bool]

    validate_task_outputs = parameter(default=True).help(
        "Determine whether dbnd task should check that all outputs exist after the task has been executed."
    )[bool]

    validate_task_outputs_on_build = parameter(default=False).help(
        "Determine whether dbnd task should check that there are no incomplete outputs before the task executes."
    )[bool]

    pipeline_band_only_check = parameter(default=False).help(
        "When checking if the pipeline is completed, check only if the band file exists, and skip the tasks."
    )[bool]

    task_complete_parallelism_level = parameter(default=1).help(
        "Set the number of threads to use when checking if tasks are already complete."
    )[int]

    pool = parameter(
        default=None, description="Determine which resource pool will be used."
    )[str]

    ######
    # Advanced Run settings (debug/workarounds)
    # run .pickle file
    always_save_pipeline = parameter(
        description="Enable always saving pipeline to pickle."
    ).value(False)
    disable_save_pipeline = parameter(description="Disable pipeline pickling.").value(
        False
    )
    donot_pickle = parameter(
        description="Do not attempt to pickle the DAG object to send over to the workers. "
        "Instead, tell the workers to run their version of the code."
    ).value(False)
    pickle_handler = parameter(
        default=None,
        description="Define a python pickle handler to be used to pickle the run's data",
    )[str]
    enable_concurent_sqlite = parameter(
        description="Enable concurrent execution with sqlite database. This should only be used for debugging purposes."
    ).value(False)

    ######
    # HEARTBEAT (process that updates on driver status every `heartbeat_interval_s`
    #
    heartbeat_interval_s = parameter(
        description="Set how often a run should send a heartbeat to the server. "
        "Setting this to -1 will disable it altogether."
    )[int]
    heartbeat_timeout_s = parameter(
        description="Set how old a run's last heartbeat can be before we consider it to have failed."
        "Setting this to -1 will disable this."
    )[int]
    heartbeat_sender_log_to_file = parameter(
        description="Enable creating a separate log file for the heartbeat sender and don't log the run process stdout."
    )[bool]

    hearbeat_disable_plugins = parameter(
        default=False, description="Disable dbnd plugins at heartbeat sub-process."
    )[bool]

    ######
    # Task/Pipeline in task Execution
    task_run_at_execution_time_enabled = parameter(
        default=True, description="Allow tasks calls during another task's execution."
    )[bool]
    task_run_at_execution_time_in_memory_outputs = parameter(
        default=False,
        description="Enable storing outputs for an inline task at execution time in memory. (do not use FileSystem)",
    )[bool]
    target_cache_on_access = parameter(
        default=True,
        description="Enable caching target values in memory during execution.",
    )[bool]

    # BOOTSTRAP
    module = parameter(
        default=None, description="Auto load this module before resolving user classes"
    )[str]

    # USER CODE TO RUN ON START
    user_configs = parameter(
        empty_default=True,
        description="Set the config used for creating tasks from user code.",
    )[List[str]]

    # user_pre_init = defined at Databand System config, dbnd_on_pre_init_context
    user_init = parameter(
        default=None,
        description="This runs in every DBND process with System configuration in place. This is called in "
        "DatabandContex after entering initialization steps.",
    )[object]
    user_driver_init = parameter(
        default=None,
        description="This runs in driver after configuration initialization. This is called from DatabandContext when "
        "entering a new context(dbnd_on_new_context)",
    )[object]

    user_code_on_fork = parameter(
        default=None,
        description="This runs in a sub process, on parallel, kubernetes, or external modes.",
    )[object]

    # PLUGINS
    plugins = parameter(
        description="Specify which plugins should be loaded on Databand context creations.",
        default=None,
    )[str]
    fix_env_on_osx = parameter(
        description="Enable adding `no_proxy=*` to environment variables, fixing issues with multiprocessing on OSX."
    )[bool]

    environments = parameter(description="Set a list of enabled environments.")[
        List[str]
    ]



# © Copyright Databand.ai, an IBM Company 2022

import base64
import logging
import os

from dbnd._core.utils.cli import required_mutually_exclusive_options
from dbnd._vendor import click


logger = logging.getLogger(__name__)


@click.command(
    help="Collect logs and debugging information for a specific DatabandRun. Creates a tarfile that can be "
    "easily sent to databand.ai customer support"
)
@required_mutually_exclusive_options("uid", "name")
@click.option("--name", "-n", help="The name of the databand run to retrieve logs for")
@click.option("--uid", "-u", help="The UUID of the databand run to retrieve logs for")
def collect_logs(name, uid):
    # Click performs parameter validation, mutually exclusive options

    output_filename, tar_file_data = send_collect_logs_api_request(name, uid)
    working_directory = os.getcwd()
    with open(os.path.join(working_directory, output_filename), "wb") as output_file:
        output_file.write(base64.b64decode(tar_file_data))
    logger.info(
        "Successfully written tarfile %s to disk! Please send it to our customer support to continue the "
        "debugging process!" % output_filename
    )


def send_collect_logs_api_request(name, uid):
    from dbnd import get_databand_context
    from dbnd._core.errors.base import DatabandApiError
    from dbnd._core.errors.friendly_error.api import couldnt_find_databand_run_in_db

    COLLECT_LOGS_ENDPOINT = "runs/collect_error_info"

    query = {}
    using_name = True
    if name:
        query["run_name"] = name
    elif uid:
        using_name = False
        query["run_uid"] = uid

    try:
        api_client = get_databand_context().databand_api_client
        response = api_client.api_request(
            endpoint=COLLECT_LOGS_ENDPOINT, method="GET", query=query, data={}
        )
        return response["name"], response["data"]
    except DatabandApiError as e:
        if using_name:
            raise couldnt_find_databand_run_in_db(query["run_name"], e)
        else:
            raise couldnt_find_databand_run_in_db(query["run_uid"], e)



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.utils.click_tzdatetime import TZAwareDateTime
from dbnd._vendor import click
from dbnd._vendor.click import get_current_context
from dbnd.api.scheduler import (
    ScheduledJobNamedTuple,
    delete_scheduled_job,
    get_scheduled_jobs,
    patch_scheduled_job,
    post_scheduled_job,
    set_scheduled_job_active,
)


SCHEDULED_JOB_HEADERS = [
    "name",
    "active",
    "cmd",
    "schedule_interval",
    "readable_schedule_interval",
    "last_job_date",
    "next_job_date",
]

SCHEDULED_JOB_VERBOSE_HEADERS = SCHEDULED_JOB_HEADERS + [
    "start_date",
    "end_date",
    "catchup",
    "depends_on_past",
    "retries",
    "create_user",
    "create_time",
    "update_user",
    "update_time",
    "extra_args",
]

datetime_formats = [
    "%Y-%m-%d",
    "%Y-%m-%dT%H:%M:%S",
    "%Y-%m-%dT%H:%M:%SZ%z",
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M:%SZ%z",
]

SPECIAL_INTERVALS = {"@never": None}


@click.group()
@click.pass_context
def schedule(ctx):
    """Manage scheduled jobs"""

    dbnd_bootstrap(enable_dbnd_run=True)

    ctx.obj = {}
    ctx.obj["headers"] = SCHEDULED_JOB_HEADERS
    from dbnd import new_dbnd_context

    new_dbnd_context(conf={"core": {"tracker": ""}}).__enter__()


@schedule.command()
@click.option(
    "--all", "-a", "include_all", is_flag=True, help="Lists all deleted scheduled jobs"
)
@click.option("--verbose", "-v", is_flag=True, help="Print extra job details")
@click.pass_context
def list(ctx, include_all, verbose):
    """List scheduled jobs"""
    include_deleted = None if include_all else False
    scheduled_jobs = get_scheduled_jobs(include_deleted=include_deleted)

    ctx.obj["headers"] = (
        SCHEDULED_JOB_VERBOSE_HEADERS if verbose else SCHEDULED_JOB_HEADERS
    )
    _click_echo_jobs(scheduled_jobs)


@schedule.command()
@click.option("--name", "-n", help="Name of the scheduled job to enable)")
def enable(name):
    """Enable scheduled job"""
    set_scheduled_job_active(name, True)

    click.echo('Scheduled job "%s" is enabled' % name)


@schedule.command()
@click.option("--name", "-n", help="Name of the scheduled job to pause)")
def pause(name):
    """Pause scheduled job"""
    set_scheduled_job_active(name, False)

    click.echo('Scheduled job "%s" paused' % name)


@schedule.command()
@click.option("--name", "-n", help="Name of the scheduled job to undelete)")
def undelete(name):
    """Un-Delete deleted scheduled job"""
    delete_scheduled_job(name, revert=True)


@schedule.command()
@click.option("--name", "-n", help="Name of the scheduled job to delete)")
@click.option("--force", "-f", is_flag=True, help="Delete without confirmation")
def delete(name, force):
    """Delete scheduled job"""
    if not force:
        to_delete = get_scheduled_jobs(name_pattern=name)
        if not to_delete:
            click.echo("no jobs found matching the given name pattern")
            return

        click.echo("the following jobs will be deleted:")
        _click_echo_jobs(to_delete)

        click.confirm("are you sure?", abort=True)

    delete_scheduled_job(name)


@schedule.command()
@click.option("--name", "-n", help="Name of the scheduled job (must be unique)")
@click.option("--cmd", "-c", help="Shell command to run")
@click.option(
    "--schedule-interval",
    "-si",
    help="Any valid cron expression or one of the presets: "
    "@once, @hourly, @daily, @weekly, @monthly, @yearly or @never. "
    + "See documentation for precise definitions of the presets",
)
@click.option("--start-date", "-s", type=TZAwareDateTime(formats=datetime_formats))
@click.option("--end-date", "-ed", type=TZAwareDateTime(formats=datetime_formats))
@click.option(
    "--catchup",
    "-cu",
    type=bool,
    help="Run scheduled past tasks that have not been run. For example: a pipeline is scheduled to run every day"
    + "but the scheduler didn't run for a week. This will cause a catchup run to be scheduled for every day missed. "
    + "Default is determined by the catchup_by_default config flag.",
)
@click.option(
    "--depends-on-past",
    "-dp",
    type=bool,
    help="Only run a job if it's previous run succeeded",
)
@click.option(
    "--retries",
    "-r",
    type=int,
    help="Number of times to retry a failed run. If unset will use the default set in the config file",
)
@click.option(
    "--update",
    "-u",
    is_flag=True,
    help="If a job with the given name already exists partially update it with the given params "
    "(unspecified params will be left as is)",
)
@click.option(
    "--extra-args",
    "-ea",
    help="The JSON struct containing the extra arguments (any additional data)",
)
def job(
    name,
    cmd,
    schedule_interval,
    start_date,
    end_date,
    catchup,
    depends_on_past,
    update,
    retries,
    extra_args,
):
    """Manage scheduled jobs"""
    from dbnd._core.utils.timezone import is_localized, make_aware

    if start_date and not is_localized(start_date):
        start_date = make_aware(start_date)
    if end_date and not is_localized(end_date):
        end_date = make_aware(end_date)
    scheduled_job = {
        "name": name,
        "cmd": cmd,
        "start_date": start_date,
        "end_date": end_date,
        "schedule_interval": _replace_special_intervals(schedule_interval),
        "catchup": catchup,
        "depends_on_past": depends_on_past,
        "retries": retries,
        "extra_args": extra_args,
    }

    if update:
        res = patch_scheduled_job(scheduled_job)
    else:
        res = post_scheduled_job(scheduled_job)

    _click_echo_jobs([res])


def _click_echo_jobs(jobs):
    from dbnd._core.utils.object_utils import tabulate_objects

    headers = get_current_context().obj["headers"]

    job_objects = [
        ScheduledJobNamedTuple(**dict(j.pop("DbndScheduledJob"), **j)) for j in jobs
    ]

    click.echo(tabulate_objects(job_objects, headers=headers, to_local_time=False))


def _replace_special_intervals(schedule_interval: str):
    return SPECIAL_INTERVALS.get(schedule_interval, schedule_interval)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from contextlib import contextmanager

from dbnd._core.log.log_preview import read_dbnd_log_preview
from dbnd._core.log.logging_utils import capture_stderr_stdout
from dbnd_run.run_settings import RunLoggingConfig
from dbnd_run.task_ctrl import _TaskRunExecutorCtrl


if typing.TYPE_CHECKING:
    from targets import FileTarget

logger = logging.getLogger(__name__)


class TaskRunOrchestrationLogManager(_TaskRunExecutorCtrl):
    def __init__(self, task_run, local_log_file, remote_log_file, extra_log_file=None):
        super(TaskRunOrchestrationLogManager, self).__init__(task_run)

        self.local_log_file: FileTarget = local_log_file

        self.extra_log_file = extra_log_file

        self.remote_log_file: FileTarget = remote_log_file

        # file handler for task log
        # if set -> we are in the context of capturing
        self._log_task_run_into_file_active = False

    @contextmanager
    def capture_task_log(self):
        log_file = self.local_log_file

        run_logging: RunLoggingConfig = self.run_executor.run_settings.run_logging
        if self._log_task_run_into_file_active:
            yield None
            return

        if (
            # there is no really any good way to find if this is runtime tracking for airflow operator
            # should be refactored as soon as possible
            self.task.task_family.endswith("_execute")
        ):
            yield None
            return

        handler = run_logging.get_task_log_file_handler(log_file)
        logger.debug("Capturing task log into '%s'", log_file)

        if not handler:
            yield None
            return
        target_logger = logging.root

        try:
            target_logger.addHandler(handler)
            self._log_task_run_into_file_active = True

            if self.task.settings.tracking_log.capture_stdout_stderr:
                with capture_stderr_stdout():
                    yield handler
            else:
                yield handler

        except Exception as task_ex:
            raise task_ex
        finally:
            try:
                target_logger.removeHandler(handler)
                handler.close()
            except Exception:
                logger.warning("Failed to close file handler for log %s", log_file)

            self._log_task_run_into_file_active = False
            self._upload_task_log_preview_to_databand(handler)

    def _upload_task_log_preview_to_databand(self, log_handler):
        try:
            if isinstance(log_handler, logging.FileHandler):
                log_body = self.read_log_body()
            else:
                logger.exception("unable to read log body for %s", self)
                return
        except Exception as read_log_err:
            logger.exception("failed to read log preview for %s:%s", self, read_log_err)
        else:
            try:
                self.task_run.tracker.save_task_run_log(log_body, self.local_log_file)
            except Exception as save_log_err:
                logger.exception(
                    "failed to save task run log preview for %s:%s", self, save_log_err
                )

        self.write_remote_log()

    def read_log_body(self):
        try:
            return read_dbnd_log_preview(self.local_log_file.path, self.extra_log_file)

        except Exception as ex:
            logger.warning(
                "Failed to read log (%s) for %s: %s",
                self.local_log_file.path,
                self.task,
                ex,
            )
            return None

    def write_remote_log(self):
        if not self.run_executor.run_settings.run_logging.remote_logging_disabled:
            return

        if self.remote_log_file is None or self.local_log_file is None:
            return

        try:
            self.remote_log_file.copy_from_local(self.local_log_file.path)

        except Exception as ex:
            logger.warning(
                "Failed to upload the log from {local_path} "
                "to the remote log at {remote_log} "
                "for task {task}: {exception}".format(
                    local_path=self.local_log_file.path,
                    remote_log=self.remote_log_file.path,
                    task=self.task,
                    exception=ex,
                )
            )



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.constants import SystemTaskName
from dbnd._core.task_ctrl.task_visualiser import (
    FormatterVerbosity,
    _f_env,
    _f_none_default,
    _TaskBannerBuilder,
)
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.utils.basics.text_banner import TextBanner
from dbnd_run.task.task import Task


class _TaskOrchestrationBannerBuilder(_TaskBannerBuilder):
    """
    Describe how to build a banner for different scenarios.
    Also describe how to build each of the columns in the banner.
    todo: can be extracted to builder and director
    """

    def __init__(
        self,
        task: "Task",
        msg: str,
        color: str,
        verbose: bool = False,
        print_task_band: bool = False,
    ):
        super().__init__(task=task, msg=msg, color=color, verbose=verbose)
        self.task: "Task" = task
        self.print_task_band = print_task_band
        self.is_driver_or_submitter = (
            self.task.task_name in SystemTaskName.driver_and_submitter
        )

    def build_orchestration_banner(
        self, task_run: "TaskRun" = None, exc_info=None
    ) -> TextBanner:
        """building the banner for orchestration scenario"""

        self.add_task_info()

        if not self.task.ctrl.should_run():
            self.add_disabled_info()

        if task_run:
            self.add_time_info(task_run)
            self.add_tracker_info(task_run)
            self.add_task_run_info(task_run)

            if task_run.external_resource_urls:
                self.add_external_resource_info(task_run)

            self.add_task_run_executor_log_info(task_run)

        all_info = self.verbosity >= FormatterVerbosity.HIGH
        self.table_director.add_params_table(
            all_params=all_info,
            param_format=True,
            param_source=True,
            param_section=all_info,
            param_default=all_info,
        )

        if self.verbosity >= FormatterVerbosity.HIGH:
            self._add_verbose_info()

        elif self.print_task_band:
            self._add_task_band_info()

        self.task._task_banner(self.banner, verbosity=self.verbosity)
        self.add_stack_and_errors(exc_info, task_run, self.verbosity)

        return self.banner.get_banner_str()

    def add_task_run_executor_log_info(self, task_run):
        log_manager = task_run.task_run_executor.log_manager
        logs = [("local", log_manager.local_log_file)]
        if log_manager.remote_log_file:
            logs.append(("remote", log_manager.remote_log_file))
        self.banner.column("LOG", self.banner.f_simple_dict(logs))

    def add_external_resource_info(self, task_run):
        self.banner.column_properties(
            "EXTERNAL", [(k, v) for k, v in task_run.external_resource_urls.items()]
        )

    def add_task_run_info(self, task_run):
        self.banner.column_properties(
            "TASK RUN",
            [
                ("task_run_uid", task_run.task_run_uid),
                ("task_run_attempt_uid", task_run.task_run_attempt_uid),
                ("state", task_run.task_run_state),
            ],
        )

    def add_task_info(self):
        from dbnd_run.task.task import DEFAULT_CLASS_VERSION

        task_params = [
            ("task_id", self.task.task_id),
            ("task_version", self.task.task_version),
        ]

        if self.task.task_class_version != DEFAULT_CLASS_VERSION:
            task_class_version = (
                _f_none_default(self.task.task_class_version, DEFAULT_CLASS_VERSION),
            )
            task_params.append(("task_class_version", task_class_version))

        task_params.extend(
            [
                ("env", self.task.task_env.task_name),
                ("env_cloud", self.task.task_env.cloud_type),
                ("env_label", _f_env(self.task.task_env.env_label)),
                ("task_target_date", self.task.task_target_date),
            ]
        )
        self.banner.column("TASK", self.banner.f_simple_dict(task_params))

    def add_disabled_info(self):
        disable = [
            ("task_enabled", self.task.task_enabled),
            ("task_enabled_in_prod", self.task.task_enabled_in_prod),
        ]

        self.banner.column("DISABLED", self.banner.f_simple_dict(disable))

    def add_time_info(self, task_run):
        time_fields = [("start", "%s" % task_run.start_time)]

        if task_run.finished_time:
            time_fields.append(("finished", "%s" % task_run.finished_time))
            task_duration = task_run.finished_time - task_run.start_time
            time_fields.append(("duration", "%s" % task_duration))

        self.banner.column("TIME", self.banner.f_simple_dict(time_fields))

    def _add_verbose_info(self):
        relations = self.task.ctrl.relations
        self.banner.new_section()
        self.banner.column("INPUTS USER", self.banner.f_io(relations.task_inputs_user))
        self.banner.column(
            "INPUTS SYSTEM", self.banner.f_struct(relations.task_inputs_system)
        )
        self.banner.new_section()

        task = self.task
        self.banner.column("TASK_BAND", task.task_band)
        self.banner.column(
            "OUTPUTS USER", self.banner.f_io(relations.task_outputs_user)
        )
        self.banner.column(
            "OUTPUTS SYSTEM", self.banner.f_struct(relations.task_outputs_system)
        )
        self.banner.write("\n")

        self.banner.column("SIGNATURE", task.task_signature_obj.signature)
        self.banner.column("SIGNATURE SOURCE", task.task_signature_obj.signature_source)
        if (
            task.task_outputs_signature_obj
            and task.task_outputs_signature_obj != task.task_signature_obj
        ):
            self.banner.column(
                "TASK OUTPUTS SIGNATURE", task.task_outputs_signature_obj.signature
            )
            self.banner.column(
                "TASK OUTPUTS SIGNATURE SOURCE",
                task.task_outputs_signature_obj.signature_source,
            )

    def _add_task_band_info(self):
        self.banner.new_section()
        self.banner.column("TASK_BAND", self.task.task_band)



# © Copyright Databand.ai, an IBM Company 2022

from typing import List

from dbnd._core.configuration.dbnd_config import DbndConfig
from dbnd._core.utils.basics.load_python_module import load_python_module


def load_user_modules(dbnd_config: DbndConfig, modules: List[str] = None) -> None:
    # loading user modules
    module_from_config = dbnd_config.get("run", "module")
    if module_from_config:
        load_python_module(module_from_config, "config file (see [databand].module)")
    if modules:
        for m in modules:
            load_python_module(m, "--module")



# © Copyright Databand.ai, an IBM Company 2022

"""
 Usage sample:

 python -m dbnd_run.airflow.scheduler.utils.generate_partitioned_dags --partitions=3 --dag-folder=dags --base-name=dbnd_dags_from_databand --template=template.py

 --dag-folder and --template can be relative or absolute path
"""
import os

from pathlib import Path
from textwrap import dedent
from typing import Optional

import click


dag_file_template = dedent(
    """# airflow will only scan files containing the text DAG or airflow. This comment performs this function

from dbnd_run.airflow.scheduler.dags_provider_from_databand import get_dags_from_databand
from dbnd_run.airflow.scheduler.utils.partitioning import partition_from_module_name

dags = get_dags_from_databand(partition=partition_from_module_name(__name__, base_filename="%base_filename%_"))
if dags:
    globals().update(dags)
"""
)


@click.command()
@click.option(
    "--partitions",
    type=click.INT,
    required=True,
    help="Total number of partitions to generate",
)
@click.option(
    "--base-name",
    default="dbnd_dags_from_databand",
    type=click.STRING,
    help="Path prefix, ex: dbnd_dags_from_databand -> dbnd_dags_from_databand_1__3.py",
)
@click.option(
    "--dag-folder",
    required=True,
    type=click.STRING,
    help="Path to AIRFLOW_HOME/dags folder",
)
@click.option("--template", type=click.STRING, help="Path custom template file")
def generate_dags_files(
    dag_folder: str, base_name: str, partitions: int, template: Optional[str]
):
    print(f"Generating {partitions} files at {dag_folder}")
    print(f"File name prefix: {base_name}")

    if not os.path.exists(dag_folder):
        print(
            f"Dag folder {dag_folder} doesn't exist. Make sure you are providing valid folder or create it before usage."
        )
        return

    existing_files = os.listdir(dag_folder)
    for dag_file in existing_files:
        if dag_file.startswith(base_name):
            raise Exception(
                f"It seems that folder already contains {base_name} files. Try to remove them first and re-run the script."
            )

    dag_folder_path = Path(dag_folder)

    files_generated = []
    dag_file_content = dag_file_template
    if template:
        with open(template, "r") as template_file:
            dag_file_content = template_file.read()

    final_content = dag_file_content.replace("%base_filename%", base_name)
    for part in range(0, partitions):
        current_path = dag_folder_path.joinpath(base_name + f"_{part}__{partitions}.py")
        files_generated.append(str(current_path))
        with open(current_path, "w") as current_file:
            current_file.write(final_content)

    print("Completed successfully.")
    print("Files generated:")
    [print(f"\t{filename}") for filename in files_generated]


if __name__ == "__main__":
    generate_dags_files()



# Vendorized from Apache Airflow
# License is Apache 2.0
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# This file has been modified by databand.ai to support dbnd orchestration.


import hashlib
import re

MAX_LABEL_LEN = 63


def make_safe_label_value(string):
    """
    Valid label values must be 63 characters or less and must be empty or begin and
    end with an alphanumeric character ([a-z0-9A-Z]) with dashes (-), underscores (_),
    dots (.), and alphanumerics between.

    If the label value is then greater than 63 chars once made safe, or differs in any
    way from the original value sent to this function, then we need to truncate to
    53chars, and append it with a unique hash.
    """
    safe_label = re.sub(r"^[^a-z0-9A-Z]*|[^a-zA-Z0-9_\-\.]|[^a-z0-9A-Z]*$", "", string)

    if len(safe_label) > MAX_LABEL_LEN or string != safe_label:
        safe_hash = hashlib.md5(string.encode()).hexdigest()[:9]
        safe_label = safe_label[: MAX_LABEL_LEN - len(safe_hash) - 1] + "-" + safe_hash

    return safe_label



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import signal
import typing
import webbrowser

from dbnd._core.constants import SystemTaskName, TaskRunState
from dbnd._core.errors import show_error_once
from dbnd._core.errors.base import DatabandSigTermError
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.task_run.task_run_meta_files import TaskRunMetaFiles
from dbnd._core.utils import seven
from dbnd._core.utils.basics.nested_context import nested
from dbnd._core.utils.basics.signal_utils import safe_signal
from dbnd._core.utils.seven import contextlib
from dbnd._core.utils.timezone import utcnow
from dbnd_run import errors
from dbnd_run.plugin.dbnd_plugins import is_plugin_enabled, pm
from dbnd_run.run_settings import LocalEnvConfig
from dbnd_run.task_ctrl import _TaskRunExecutorCtrl
from dbnd_run.task_ctrl.task_run_orchestration_logging import (
    TaskRunOrchestrationLogManager,
)
from dbnd_run.task_ctrl.task_run_sync_local import TaskRunLocalSyncer
from dbnd_run.task_ctrl.task_sync_ctrl import TaskSyncCtrl
from dbnd_run.task_ctrl.task_validator import TaskValidator
from targets import FileTarget, target
from targets.target_config import TargetConfig


if typing.TYPE_CHECKING:
    from dbnd import Task

logger = logging.getLogger(__name__)


class TaskRunExecutor(_TaskRunExecutorCtrl):
    def __init__(self, task_run, run_executor, task_engine):
        super(TaskRunExecutor, self).__init__(task_run=task_run)

        self.deploy = TaskSyncCtrl(task_run=task_run)
        self.sync_local = TaskRunLocalSyncer(task_run=task_run)

        self.task_validator = TaskValidator(task_run=task_run)

        # custom per task engine , or just use one from global env
        self.task_engine = task_engine
        dbnd_local_root = (
            self.task_engine.dbnd_local_root or run_executor.env.dbnd_local_root
        )
        self.local_task_run_root = (
            dbnd_local_root.folder(run_executor.run_folder_prefix)
            .folder("tasks")
            .folder(self.task.task_id)
        )

        # trying to find if we should use attempt_uid that been set from external process.
        # if so - the attempt_uid is uniquely for this task_run_attempt, and that why we pop.
        task_run = self.task_run
        self.attempt_folder = self.task._meta_output.folder(
            "attempt_%s_%s"
            % (task_run.attempt_number, self.task_run.task_run_attempt_uid),
            extension=None,
        )
        self.attempt_folder_local = self.local_task_run_root.folder(
            "attempt_%s_%s" % (task_run.attempt_number, self.task_run_attempt_uid),
            extension=None,
        )
        self.attemp_folder_local_cache = self.attempt_folder_local.folder("cache")
        self.meta_files = TaskRunMetaFiles(self.attempt_folder)

        self.local_log_file: FileTarget = self.local_task_run_root.partition(
            name="%s.log" % task_run.attempt_number
        )

        self.remote_log_file = None
        if (
            not isinstance(self.task.task_env, LocalEnvConfig)
            and not self.run_executor.run_settings.run_logging.remote_logging_disabled
        ):
            self.remote_log_file: FileTarget = self.attempt_folder.partition(
                name=str(task_run.attempt_number),
                config=TargetConfig().as_file().txt,
                extension=".log",
            )

        self.log_manager = TaskRunOrchestrationLogManager(
            task_run=task_run,
            remote_log_file=self.remote_log_file,
            local_log_file=self.local_log_file,
        )

    def task_run_attempt_file(self, *path):
        return target(self.attempt_folder, *path)

    def execute(self, airflow_context=None, allow_resubmit=True, handle_sigterm=True):
        task_run = self.task_run
        run_executor = self.run_executor
        run_config = run_executor.run_config
        task = self.task  # type: Task
        task_engine = self.task_engine
        if allow_resubmit and task_engine._should_wrap_with_submit_task(task_run):
            args = task_engine.dbnd_executable + [
                "execute",
                "--dbnd-run",
                str(run_executor.driver_dump),
                "task_execute",
                "--task-id",
                task_run.task.task_id,
            ]
            submit_task = task_engine.submit_to_engine_task(
                env=task.task_env, task_name=SystemTaskName.task_submit, args=args
            )
            submit_task.descendants.add_child(task.task_id)
            if run_config.open_web_tracker_in_browser:
                webbrowser.open_new_tab(task_run.task_tracker_url)
            run_executor.run_task_at_execution_time(submit_task)
            return

        with self.task_run.task_run_track_execute(
            capture_log=False
        ) as tracking_context, self.task_run_execute_context(
            handle_sigterm=handle_sigterm
        ) as execute_context:

            if run_executor.is_killed():
                raise errors.task_execution.databand_context_killed(
                    "task.execute_start of %s" % task
                )
            try:
                self.task_env.prepare_env()
                if run_config.skip_completed_on_run and task._complete():
                    task_run.set_task_reused()
                    return
                task_run.set_task_run_state(state=TaskRunState.RUNNING)

                if not self.task.ctrl.should_run():
                    self.task.ctrl.validator.find_and_raise_missing_inputs()

                if run_config.validate_task_inputs:
                    self.task_validator.validate_task_inputs()

                try:
                    result = self.task._task_submit()
                    self.ctrl.save_task_band()
                    if run_config.validate_task_outputs:
                        self.task_validator.validate_task_is_complete()
                finally:
                    self.task_run.finished_time = utcnow()

                task_run.set_task_run_state(TaskRunState.SUCCESS)
                run_executor.cleanup_after_task_run(task)

                return result
            except DatabandSigTermError as ex:
                logger.error(
                    "Sig TERM! Killing the task '%s' via task.on_kill()",
                    task_run.task.task_id,
                )
                run_executor._internal_kill()

                error = TaskRunError.build_from_ex(ex, task_run)
                try:
                    task.on_kill()
                except Exception:
                    logger.exception("Failed to kill task on user keyboard interrupt")
                task_run.set_task_run_state(TaskRunState.CANCELLED, error=error)
                raise
            except KeyboardInterrupt as ex:
                logger.warning(
                    "User Interrupt! Killing the task %s", task_run.task.task_id
                )
                error = TaskRunError.build_from_ex(ex, task_run)
                try:
                    if task._conf_confirm_on_kill_msg:
                        from dbnd._vendor import click

                        if click.confirm(task._conf_confirm_on_kill_msg, default=True):
                            task.on_kill()
                        else:
                            logger.warning(
                                "Task is not killed accordingly to user input!"
                            )
                    else:
                        task.on_kill()
                except Exception:
                    logger.exception("Failed to kill task on user keyboard interrupt")
                task_run.set_task_run_state(TaskRunState.CANCELLED, error=error)
                run_executor._internal_kill()
                raise
            except SystemExit as ex:
                error = TaskRunError.build_from_ex(ex, task_run)
                task_run.set_task_run_state(TaskRunState.CANCELLED, error=error)
                raise errors.task_execution.system_exit_at_task_run(task, ex)
            except Exception as ex:
                error = TaskRunError.build_from_ex(ex, task_run)
                task_run.set_task_run_state(TaskRunState.FAILED, error=error)
                show_error_once.set_shown(ex)
                raise
            finally:
                task_run.airflow_context = None

    @contextlib.contextmanager
    def task_run_execute_context(self, handle_sigterm=True):
        ctx_managers = []

        if handle_sigterm:
            ctx_managers.append(handle_sigterm_at_dbnd_task_run())

        if self.run_executor.run_settings.run_logging.capture_task_run_log:
            ctx_managers.append(self.log_manager.capture_task_log())

        ctx_managers.extend(pm.hook.dbnd_task_run_context(task_run=self.task_run))

        with nested(*ctx_managers):
            yield

    def __getstate__(self):
        d = self.__dict__.copy()
        if "airflow_context" in d:
            # airflow context contains "auto generated code" that can not be pickled (Vars class)
            # we don't need to pickle it as we pickle DAGs separately
            d = self.__dict__.copy()
            del d["airflow_context"]
        return d


@seven.contextlib.contextmanager
def handle_sigterm_at_dbnd_task_run():
    def signal_handler(signum, frame):
        logger.info("Task runner received signal. PID: %s. Exiting...", os.getpid())
        if is_plugin_enabled("dbnd-docker") and is_plugin_enabled("dbnd-airflow"):
            from dbnd_docker.kubernetes.kubernetes_engine_config import (
                ENV_DBND_POD_NAME,
            )

            if ENV_DBND_POD_NAME in os.environ:
                # We are running inside cluster
                # We should log all events on sigterm for debugging when running inside cluster
                from dbnd_run.airflow.dbnd_airflow_contrib.kubernetes_metrics_logger import (
                    log_pod_events_on_sigterm,
                )

                log_pod_events_on_sigterm(frame)

        raise DatabandSigTermError(
            "Task received signal", help_msg="Probably the job was canceled"
        )

    original_sigterm_signal = None
    try:
        original_sigterm_signal = safe_signal(signal.SIGTERM, signal_handler)
        yield
    finally:
        if original_sigterm_signal:
            safe_signal(signal.SIGTERM, original_sigterm_signal)



# © Copyright Databand.ai, an IBM Company 2022

import abc

from dbnd._core.constants import TaskType
from dbnd_run.current import try_get_run_executor
from dbnd_run.task.task import Task


class PipelineTask(Task):
    """
    Use for tasks that only wrap other tasks and that by definition are done if all their requirements exist.

    Example::

        class PrepareData(PipelineTask):
            data = parameter.data
            prepared_data = output.csv.data

            def band(self):
                self.prepared_data = gather_data.dbnd_run()
    """

    _conf__task_type_name = TaskType.pipeline

    def _task_run(self):
        """We override, as we don't want to automatically load deferred inputs as we do it in regular task."""
        result = self.run()
        return result

    @abc.abstractmethod
    def band(self):
        """
        This is the method you should override while using PipelineTask.

        we need to implement result of this function to be "output" of the task (task_output, and vars)
        Your Pipeline.band() call should have one or more tasks wired one into another.
        See examples!"
        """
        return

    def _complete(self):
        if self.task_band:
            if not self.task_band.exists():
                return False
            # With very large pipelines, checking all tasks might take a very long time
            # so we might want to assume that if the band exist, probably all outputs also exist
            re = try_get_run_executor()
            if re and re.run_config.pipeline_band_only_check:
                return True
        return super(PipelineTask, self)._complete()


PipelineTask.task_definition.hidden = True



# © Copyright Databand.ai, an IBM Company 2022
from __future__ import absolute_import

from typing import TypeVar, Union

from dbnd import dbnd_run_cmd, output, parameter
from dbnd._core.run.databand_run import DatabandRun
from dbnd_run.task.python_task import PythonTask


T = TypeVar("T")


def run_locally__raises(expected_exception, args, conf=None):
    import pytest

    with pytest.raises(expected_exception):
        dbnd_run_cmd(args)


def assert_run_task(task):  # type: (Union[T, Task]) -> Union[T, Task]
    task.dbnd_run()
    assert task._complete()
    return task


def assert_run(task):  # type: (Union[T, Task]) -> DatabandRun
    run = task.dbnd_run()
    assert task._complete()
    return run


class TTask(PythonTask):
    t_param = parameter.value("1")
    t_output = output.data

    def run(self):
        self.t_output.write("%s" % self.t_param)



# © Copyright Databand.ai, an IBM Company 2022

from os import path

import setuptools

from setuptools.config import read_configuration


BASE_PATH = path.dirname(__file__)
CFG_PATH = path.join(BASE_PATH, "setup.cfg")

config = read_configuration(CFG_PATH)
version = config["metadata"]["version"]

setuptools.setup(
    name="dbnd-run",
    package_dir={"": "src"},
    install_requires=["dbnd==" + version, "packaging"],
    # Only for orchestration, tracking users should install Airflow manually before DBND
    # The best way to install airflow is manually with constraints beforehand.
    # For example:
    # pip install apache-airflow  --constraint "https://raw.githubusercontent.com/apache/airflow/constraints-${AIRFLOW_VERSION}/constraints-{PYTHON_VERSION}.txt
    extras_require={
        "tests": [
            # # airflow support
            "pytest==6.2.5",
            "coverage==7.0.1",
            "pytest-cov==3.0.0",
            "boto3",
            "mock",
            "sh",
        ],
        "test-pandas": ["numpy<1.23", "pandas<2.0.0,>=0.17.1"],
        "test-providers": [
            "openpyxl==2.6.4",
            "numpy<1.23",
            "pandas<2.0.0,>=0.17.1",
            'matplotlib==3.3.0;python_version<"3.8"',
            'matplotlib==3.6.2;python_version>="3.8"',
            "tables==3.7.0",
            "feather-format",
            "pyarrow",
        ],
    },
    entry_points={
        "airflow.plugins": [
            "dbnd_webserver_plugin = dbnd_run.airflow.plugins.dbnd_airflow_webserver_plugin:DatabandAirflowWebserverPlugin"
        ]
    },
)



# © Copyright Databand.ai, an IBM Company 2022

import typing

from airflow import DAG
from airflow.jobs.backfill_job import BackfillJob
from airflow.timetables.base import DagRunInfo
from airflow.utils.db import provide_session
from sqlalchemy.orm.session import Session


if typing.TYPE_CHECKING:
    from dbnd_run.airflow.dbnd_task_executor.dbnd_task_executor_via_airflow import (
        AirflowTaskExecutor,
    )


class SingleDagRunJob(BackfillJob):
    __mapper_args__ = {"polymorphic_identity": "BackfillJob"}

    KEY_ERROR_ON_RERUN_RETRY = 100

    def __getstate__(self):
        state = self.__dict__.copy()
        # Don't pickle baz
        del state["dag"]
        return state

    #
    # def _execute(self, session=None):
    #
    #     return super()._execute(session=session)
    #
    #
    # def heartbeat(self, *args, **kwargs):
    #     # sync the attempt with the retries
    #     return super().heartbeat( *args, **kwargs)

    @property
    def dbnd_airflow_executor(self) -> "AirflowTaskExecutor":
        # loopback to Databand executor
        from dbnd_run.current import get_run_executor

        return get_run_executor().task_executor

    @provide_session
    def _get_dag_run(self, dagrun_info: DagRunInfo, dag: DAG, session: Session = None):
        return self.dbnd_airflow_executor.dag_run

    @provide_session
    def _execute_dagruns(
        self, dagrun_infos, ti_status, executor, pickle_id, start_date, session=None
    ):
        # we need to provide pickle_id
        for i in range(self.KEY_ERROR_ON_RERUN_RETRY):
            try:
                super()._execute_dagruns(
                    dagrun_infos=dagrun_infos,
                    ti_status=ti_status,
                    executor=executor,
                    pickle_id=self.dag.pickle_id,  # THIS IS INJECTION, we always have pickle ID!
                    start_date=start_date,
                    session=session,
                )

                # on success - return, otherwise we are retrying
                return
            except KeyError as ex:
                # Fix the error from try_number racing issue: KeyError TaskInstanceKey(dag_id=.. )
                # https://github.com/apache/airflow/pull/30669
                # https://github.com/apache/airflow/issues/13322
                # It has been fixed in 2.6.2 (may be)
                # the last fix is still not released.
                #  File  airflow/jobs/backfill_job.py", line 631, in _process_backfill_task_instances  self._update_counters(ti_status=ti_status, session=session)
                #  File "airflow/jobs/backfill_job.py", line 182, in _update_counters   ti_status.running.pop(reduced_key)
                #
                # we will try to handle it via skipping the error, so in the next iteration the task is going to be
                # checked and proceed
                if i == self.KEY_ERROR_ON_RERUN_RETRY - 1:
                    self.log.warning(
                        "KeyError in _update_counters: failed to retry, continue (most likely fail with failed jobs check)"
                    )
                    raise

                self.log.warning(
                    "KeyError in _update_counters: skipping this iteration, "
                    "the issue should be resolved in the next iteration (retry %s out of %s retries): %s",
                    i,
                    self.KEY_ERROR_ON_RERUN_RETRY,
                    ex,
                )

    def _log_progress(self, ti_status):
        self.dbnd_airflow_executor.handle_process_dag_task_instanced_iteration(
            ti_status=ti_status
        )
        super()._log_progress(ti_status=ti_status)



# © Copyright Databand.ai, an IBM Company 2022

import typing

from typing import Optional


if typing.TYPE_CHECKING:
    from dbnd_run.run_executor.run_executor import RunExecutor


def get_run_executor():
    # type: () -> RunExecutor
    """Returns current Task/Pipeline/Flow instance."""
    from dbnd_run.run_executor.run_executor import RunExecutor as _RunExecutor

    v = _RunExecutor.get_instance()
    return v


def try_get_run_executor():
    # type: () -> Optional[RunExecutor]
    from dbnd_run.run_executor.run_executor import RunExecutor as _RunExecutor

    if _RunExecutor.has_instance():
        return get_run_executor()
    return None


def is_killed():
    run_executor = try_get_run_executor()
    return run_executor and run_executor.is_killed()


def cancel_current_run(message=None):
    """Kills a run's execution from within the execution."""
    run_executor = try_get_run_executor()
    return run_executor.kill_run(message)



"""
Vendorized from SQLAlchemy-Utils version 0.36.1
License is BSD

Copyright (c) 2012, Konsta Vesterinen

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of the contributors may not be used to endorse or promote products
  derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
from copy import copy

import sqlalchemy as sa
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import OperationalError, ProgrammingError

import logging
logger= logging.getLogger(__name__)

def _sa_url_set(url, **kwargs):
    try:
        url = url.set(**kwargs)
    except AttributeError:
        # SQLAlchemy <= 1.3
        for key, value in kwargs.items():
            setattr(url, key, value)

    return url


def database_exists(url):
    """Check if a database exists.

    :param url: A SQLAlchemy engine URL.

    Performs backend-specific testing to quickly determine if a database
    exists on the server. ::

        database_exists('postgresql://postgres@localhost/name')  #=> False
        create_database('postgresql://postgres@localhost/name')
        database_exists('postgresql://postgres@localhost/name')  #=> True

    Supports checking against a constructed URL as well. ::

        engine = create_engine('postgresql://postgres@localhost/name')
        database_exists(engine.url)  #=> False
        create_database(engine.url)
        database_exists(engine.url)  #=> True

    """

    def get_scalar_result(engine, sql):
        result_proxy = engine.execute(sql)
        result = result_proxy.scalar()
        result_proxy.close()
        engine.dispose()
        return result

    def sqlite_file_exists(database):
        if not os.path.isfile(database) or os.path.getsize(database) < 100:
            return False

        with open(database, "rb") as f:
            header = f.read(100)

        return header[:16] == b"SQLite format 3\x00"

    url = copy(make_url(url))
    database = url.database
    try:
        if url.drivername.startswith("postgres"):
            url = _sa_url_set(url, database="postgres")
        else:
            url = _sa_url_set(url, database=None)
    except Exception as ex:
        logger.warning("Failed to parse URL due to SQLAlchemy version, skipping DB check..: %s", ex)
        return True

    engine = sa.create_engine(url)

    if engine.dialect.name == "postgresql":
        text = "SELECT 1 FROM pg_database WHERE datname='%s'" % database
        return bool(get_scalar_result(engine, text))

    elif engine.dialect.name == "mysql":
        text = (
            "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
            "WHERE SCHEMA_NAME = '%s'" % database
        )
        return bool(get_scalar_result(engine, text))

    elif engine.dialect.name == "sqlite":
        if database:
            return database == ":memory:" or sqlite_file_exists(database)
        else:
            # The default SQLAlchemy database is in memory,
            # and :memory is not required, thus we should support that use-case
            return True

    else:
        engine.dispose()
        engine = None
        text = "SELECT 1"
        try:
            url.database = database
            engine = sa.create_engine(url)
            result = engine.execute(text)
            result.close()
            return True

        except (ProgrammingError, OperationalError):
            return False
        finally:
            if engine is not None:
                engine.dispose()



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import logging
import typing

from airflow.utils.state import State

from dbnd._core.constants import TaskRunState, UpdateSource
from dbnd._core.current import get_databand_run
from dbnd._core.task_run.task_run import TaskRun
from dbnd_run.airflow.compat import AIRFLOW_VERSION_2
from dbnd_run.airflow.compat.airflow_multi_version_shim import is_task_instance_finished
from dbnd_run.airflow.config import AirflowConfig
from dbnd_run.current import get_run_executor


if AIRFLOW_VERSION_2:
    from airflow.ti_deps.dependencies_states import RUNNABLE_STATES
else:
    from airflow.ti_deps.dep_context import RUNNABLE_STATES

if typing.TYPE_CHECKING:
    from airflow.models import TaskInstance

logger = logging.getLogger(__name__)

SCHEDULED_OR_RUNNABLE = RUNNABLE_STATES.union({State.SCHEDULED})


def sync_task_run_attempts_retries(ti_status):
    databand_run = get_databand_run()
    for dag_run in ti_status.active_runs:
        # there is only one dag run, however, it's easier to iterate this way
        for ti in dag_run.get_task_instances():
            task_run = databand_run.get_task_run_by_af_id(ti.task_id)  # type: TaskRun
            if not task_run:
                continue
            # looking for retry tasks
            af_task_try_number = _get_af_task_try_number(ti)

            if task_run.attempt_number < af_task_try_number:
                logger.info(
                    "Found a new attempt for task %60s (dbnd=%s - af=%s) in Airflow DB(might come from Pod/Scheduler). Submitting to Databand.",
                    ti.task_id,
                    task_run.attempt_number,
                    af_task_try_number,
                )
                # update in memory object with new attempt number
                from dbnd_run.task_ctrl.task_run_executor import TaskRunExecutor

                task_run.set_task_run_attempt(af_task_try_number)
                task_run.task_run_executor = TaskRunExecutor(
                    task_run,
                    run_executor=task_run.task_run_executor.run_executor,
                    task_engine=task_run.task_run_executor.task_engine,
                )

                # sync the tracker with the new task_run_attempt
                databand_run.tracker.tracking_store.add_task_runs(
                    run=databand_run, task_runs=[task_run]
                )
                report_airflow_task_instance(ti.dag_id, ti.execution_date, [task_run])
            elif task_run.attempt_number > af_task_try_number:
                logger.info(
                    "Found a task run in Airflow DB with AF attempt lower than DBND for task %60s (dbnd=%s - af=%s), skipping",
                    ti.task_id,
                    task_run.attempt_number,
                    af_task_try_number,
                )


def update_databand_task_run_states(dagrun):
    """
    Sync states between DBND and Airflow
    we need to sync state into Tracker,
    if we use "remote" executors (parallel/k8s) we need to copy state into
    current process (scheduler)
    """

    # this is the only state we want to propogate into Databand
    # all other state changes are managed by databand itself by it's own state machine
    run_executor = get_run_executor()
    databand_run = run_executor.run

    task_runs = []

    # sync all states

    # These tasks need special treatment because Airflow doesn't manage sub-pipelines
    #   for this, we need to process failures in child tasks first
    #   and decide if the parent sub-pipeline has failed
    upstream_failed_tasks: typing.List[TaskInstance] = []

    for ti in dagrun.get_task_instances():
        task_run = databand_run.get_task_run_by_af_id(ti.task_id)  # type: TaskRun
        if not task_run:
            continue

        # UPSTREAM FAILED tasks are not going to "run" , so no code will update their state
        if (
            ti.state == State.UPSTREAM_FAILED
            and task_run.task_run_state != TaskRunState.UPSTREAM_FAILED
        ):
            upstream_failed_tasks.append(ti)

        # update only in memory state
        if (
            ti.state == State.SUCCESS
            and task_run.task_run_state != TaskRunState.SUCCESS
        ):
            task_run.set_task_run_state(TaskRunState.SUCCESS, track=False)
        if ti.state == State.FAILED and task_run.task_run_state != TaskRunState.FAILED:
            task_run.set_task_run_state(TaskRunState.FAILED, track=False)

    # process them at the last step, when we have knowledge about the child tasks
    for ti in upstream_failed_tasks:
        task_run: TaskRun = databand_run.get_task_run_by_af_id(ti.task_id)

        state = run_executor.get_upstream_failed_task_run_state(task_run)
        logger.info("Setting %s to %s", task_run.task.task_id, state)
        task_run.set_task_run_state(state, track=False)
        task_runs.append(task_run)

    # optimization to write all updates in batch
    if task_runs:
        databand_run.tracker.set_task_run_states(task_runs)


def _get_af_task_try_number(af_task_instance):
    """
    Return the expected try_number from airflow's TaskInstance

    Airflow's TaskInstance have two attributes - `_try_number` and `try_number` here is the change in flow for those two:
    State                                   _try_number     try_number      expected
    ================================================================================
    pre-run (scheduled/submitted/queued)    0               1               1
    running                                 1               1               1
    finished (success/fail/...)             1               2               1
    pre-rerun (scheduled/submitted/queued)  1               2               2
    rerunning                               2               2               2
    finished (success/fail/...)             2               3               2
    """
    if is_task_instance_finished(af_task_instance.state):
        return af_task_instance.try_number - 1

    return af_task_instance.try_number


def report_airflow_task_instance(
    dag_id, execution_date, task_runs, airflow_config=None
):
    # type: (str, datetime, List[TaskRun], Optional[AirflowConfig]) ->None
    """
    report the relevant airflow task instances to the given task runs
    """
    from dbnd.api.tracking_api import AirflowTaskInfo

    af_instances = []
    for task_run in task_runs:
        if not task_run.is_reused:
            # we build airflow infos only for not reused tasks
            af_instance = AirflowTaskInfo(
                execution_date=execution_date,
                dag_id=dag_id,
                task_id=task_run.task_af_id,
                task_run_attempt_uid=task_run.task_run_attempt_uid,
            )
            af_instances.append(af_instance)

    if airflow_config is None:
        airflow_config = AirflowConfig.from_databand_context()

    if af_instances and airflow_config.webserver_url:
        first_task_run = task_runs[0]
        first_task_run.tracker.tracking_store.save_airflow_task_infos(
            airflow_task_infos=af_instances,
            source=UpdateSource.dbnd,
            base_url=airflow_config.webserver_url,
        )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from datetime import datetime
from typing import List, Optional, Tuple

from airflow import DAG
from airflow.settings import TIMEZONE as AIRFLOW_TIMEZONE

from dbnd import set_orchestration_mode
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.configuration.environ_config import ENV_DBND_DISABLE_SCHEDULED_DAGS_LOAD
from dbnd._core.utils.basics.environ_utils import environ_enabled
from dbnd._core.utils.string_utils import clean_job_name
from dbnd._core.utils.timezone import convert_to_utc
from dbnd.api.scheduler import get_scheduled_jobs
from dbnd_run.airflow.scheduler.scheduler_dags_provider import DbndSchedulerOperator


logger = logging.getLogger(__name__)


class DbndSchedulerDBDagsProvider(object):
    # by default only run the file sync if we are in the scheduler (and not the webserver)
    def __init__(
        self, default_retries: int, custom_operator_class: Optional[type] = None
    ):
        self.custom_operator_class = custom_operator_class
        self.default_retries = default_retries

    def get_dags(self, partition: Tuple[int, int] = None):  # type: () -> List[DAG]
        logger.debug("about to get scheduler job dags from dbnd db")
        dags = []
        jobs = get_scheduled_jobs(partition=partition)
        validated_jobs = [
            s["DbndScheduledJob"]
            for s in jobs
            if not s["DbndScheduledJob"].get("validation_errors")
        ]

        for job in validated_jobs:
            if "schedule_interval" not in job:
                continue
            dag = self.job_to_dag(job)
            dags.append(dag)
        return dags

    def job_to_dag(self, job):  # type: (dict) -> Union[DAG, None]

        # convert_to_utc usage might be dangerous, as there is the same function at airflow
        # however, that one use pendulum not from _vendorized

        default_args = {}
        if job.get("depends_on_past"):
            default_args["depends_on_past"] = job.get("depends_on_past")

        start_date = self._to_airflow_datetime(job.get("start_date"))
        if start_date:
            default_args["start_day"] = start_date

        if job.get("end_date"):
            default_args["end_date"] = self._to_airflow_datetime(job.get("end_date"))

        if job.get("owner"):
            default_args["owner"] = job.get("create_user")

        job_name = clean_job_name(job["name"])
        dag = DAG(
            "%s" % job_name,
            start_date=start_date,
            default_args=default_args,
            schedule_interval=job.get("schedule_interval", None),
            catchup=job.get("catchup", False),
        )

        custom_operator_class = self.custom_operator_class or DbndSchedulerOperator
        custom_operator_class(
            scheduled_cmd=job["cmd"],
            scheduled_job_name=job_name,
            extra_args=job.get("extra_args", None),
            with_name=False,
            scheduled_job_uid=job.get("uid", None),
            shell=config.getboolean("scheduler", "shell_cmd"),
            task_id="launcher",
            dag=dag,
            retries=job.get("retries") or self.default_retries,
        )

        return dag

    def _to_airflow_datetime(self, dt: datetime) -> datetime:
        converted = convert_to_utc(dt)
        if converted is not None:
            converted = converted.replace(tzinfo=AIRFLOW_TIMEZONE)

        return converted


def get_dags_from_databand(
    custom_operator_class: Optional[type] = None, partition: Tuple[int, int] = None
):
    if environ_enabled(ENV_DBND_DISABLE_SCHEDULED_DAGS_LOAD):
        return None
    from dbnd._core.errors.base import DatabandApiError, DatabandConnectionException

    try:

        # let be sure that we are loaded
        set_orchestration_mode()
        config.load_system_configs()
        # we don't want to run the whole bootstrap

        if not config.get("core", "databand_url"):
            return {}

        default_retries = config.getint("scheduler", "default_retries")

        dags = DbndSchedulerDBDagsProvider(
            default_retries=default_retries, custom_operator_class=custom_operator_class
        ).get_dags(partition=partition)

        logger.info("providing %s dags from scheduled jobs" % len(dags))
        return {dag.dag_id: dag for dag in dags}
    except (DatabandConnectionException, DatabandApiError) as e:
        logger.error(str(e))
        raise e
    except Exception as e:
        logging.exception("Failed to get dags form databand server")
        raise e



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Type

from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    _ParameterKind,
)
from dbnd._core.parameter.parameter_value import ParameterFilters
from dbnd._vendor._marshmallow.compat import urlparse
from dbnd_run.task_ctrl import _TaskRunExecutorCtrl
from targets import DbndLocalFileMetadataRegistry, FileTarget, target
from targets.caching import get_or_create_folder_in_dir
from targets.multi_target import MultiTarget


logger = logging.getLogger(__name__)

LOCAL_SYNC_CACHE_NAME = "local_sync_cache"


class TaskRunLocalSyncer(_TaskRunExecutorCtrl):
    """
    This ctrl is in charge of syncing all required files/folders locally for execution (direct read/write)
    """

    def __init__(self, task_run):
        super(TaskRunLocalSyncer, self).__init__(task_run=task_run)
        self.inputs_to_sync = []
        self.outputs_to_sync = []
        self.local_sync_root = self.task_env.dbnd_local_root.partition("local_sync")

        for p_def, p_val in self.task._params.get_params_with_value(
            ParameterFilters.USER
        ):
            if (
                isinstance(p_val, FileTarget)
                and p_val.config.require_local_access
                and not p_val.fs.local
            ):

                self._sync_input_or_output(p_def, p_val)

            elif isinstance(p_val, MultiTarget):
                self._sync_input_or_output(p_def, p_val)

    def _local_cache_target(self, target_):
        # type: (FileTarget) -> FileTarget
        # urlparse is needed to strip filesystem prefixes like `s3://`
        path = urlparse.urlparse(target_.path)
        path = path.netloc + path.path
        path = path.lstrip("/")
        local_sync_cache_path = get_or_create_folder_in_dir(
            LOCAL_SYNC_CACHE_NAME, self.task.task_env.dbnd_local_root.path
        )
        return target(local_sync_cache_path, path, config=target_.config)

    def _local_cache_multitarget(self, multitarget):
        return MultiTarget(
            [
                self._local_cache_target(target_)
                if target_.config.require_local_access and not target_.fs.local
                else target_
                for target_ in multitarget.targets
            ]
        )

    def _sync_input_or_output(self, param_definition, old_target):
        # type: (ParameterDefinition, Type[DataTarget], Type[DataTarget]) -> None
        if param_definition.kind == _ParameterKind.task_output:
            # Output should be substituted for local path and synced post execution
            self.outputs_to_sync.append((param_definition, old_target))

        else:
            self.inputs_to_sync.append((param_definition, old_target))

    def sync_pre_execute(self):
        for p_def, remote_target in self.inputs_to_sync:
            # Input should be synced to local path and substituted
            try:
                if isinstance(remote_target, MultiTarget):
                    local_target = self._local_cache_multitarget(remote_target)

                    logger.info(
                        "Downloading  %s %s to %s", p_def, remote_target, local_target
                    )

                    for remote_subtarget, local_subtarget in zip(
                        remote_target.targets, local_target.targets
                    ):
                        if (
                            remote_subtarget.config.require_local_access
                            and not remote_subtarget.fs.local
                        ):
                            self._sync_remote_to_local(
                                remote_subtarget, local_subtarget
                            )
                else:
                    # Target requires local access, it points to a remote path that must be synced-to from a local path
                    local_target = self._local_cache_target(remote_target)

                    logger.info(
                        "Downloading  %s %s to %s", p_def, remote_target, local_target
                    )

                    self._sync_remote_to_local(remote_target, local_target)

            except Exception:
                logger.exception(
                    "Failed to create local cache for %s %s", p_def, remote_target
                )
                raise

            setattr(self.task, p_def.name, local_target)

        if self.inputs_to_sync:
            logger.info(
                "All required task inputs are downloaded to %s",
                self.task_run.task_run_executor.attemp_folder_local_cache,
            )

        for p_def, remote_target in self.outputs_to_sync:
            if isinstance(remote_target, MultiTarget):
                local_target = self._local_cache_multitarget(remote_target)
                for remote_subtarget, local_subtarget in zip(
                    remote_target.targets, local_target.targets
                ):
                    if (
                        remote_subtarget.config.require_local_access
                        and not remote_subtarget.fs.local
                    ):
                        local_subtarget.mkdir_parent()
            else:
                local_target = self._local_cache_target(remote_target)
                local_target.mkdir_parent()
            # Output should be substituted for local path and synced post execution
            setattr(self.task, p_def.name, local_target)

    def sync_post_execute(self):
        for p_def, remote_target in self.inputs_to_sync:
            setattr(self.task, p_def.name, remote_target)

        for p_def, remote_target in self.outputs_to_sync:
            try:
                if isinstance(remote_target, MultiTarget):
                    local_target = self._local_cache_multitarget(remote_target)

                    logger.info(
                        "Uploading  %s %s from %s", p_def, remote_target, local_target
                    )

                    for remote_subtarget, local_subtarget in zip(
                        remote_target.targets, local_target.targets
                    ):
                        if (
                            remote_subtarget.config.require_local_access
                            and not remote_subtarget.fs.local
                        ):
                            remote_subtarget.copy_from_local(
                                local_path=local_subtarget.path
                            )
                            if remote_subtarget.config.flag:
                                remote_subtarget.mark_success()

                else:
                    local_target = self._local_cache_target(remote_target)

                    logger.info(
                        "Uploading  %s %s from %s", p_def, remote_target, local_target
                    )

                    remote_target.copy_from_local(local_path=local_target.path)
                    if remote_target.config.flag:
                        remote_target.mark_success()
            except Exception:
                logger.exception(
                    "Failed to upload task output %s %s from %s",
                    p_def,
                    remote_target,
                    local_target,
                )
                raise
            setattr(self.task, p_def.name, remote_target)

    def _sync_remote_to_local(self, remote_target, local_target):
        # type: (FileTarget, FileTarget) -> None
        # Use DbndLocalFileMetadataRegistry to sync inputs only when necessary
        dbnd_meta_cache = DbndLocalFileMetadataRegistry.get_or_create(local_target)

        # When syncing remote to local -> We can't use MD5, so use TTL instead
        if dbnd_meta_cache.expired or not local_target.exists():
            # If TTL is invalid, or local file doesn't exist -> Sync to local
            local_target.mkdir_parent()

            with local_target.tmp() as tmp_local_path:
                remote_target.download(tmp_local_path)

        DbndLocalFileMetadataRegistry.refresh(local_target)



# © Copyright Databand.ai, an IBM Company 2022

import typing

from dbnd._vendor import pluggy


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run import TaskRun


hookspec = pluggy.HookspecMarker("dbnd")


@hookspec
def dbnd_setup_plugin():
    """Called right after plugin is imported"""


@hookspec
def dbnd_on_pre_init_context(ctx):
    """Called from DatabandContext before entering a new context"""


@hookspec
def dbnd_on_new_context(ctx):
    """Called from DatabandContext when entering a new context"""


@hookspec
def dbnd_on_existing_context(ctx):
    """Called from DatabandContext when entering an existing context"""


@hookspec
def dbnd_post_enter_context(ctx):
    """Called from DatabandContext after enter initialization steps"""


@hookspec
def dbnd_on_exit_context(ctx):
    """Called from DatabandContext when exiting"""


@hookspec
def dbnd_setup_unittest():
    """Called when running in test mode"""


@hookspec
def dbnd_task_run_context(task_run):
    # type: (TaskRun)-> Context
    """Using this context when running task_run_executor"""


@hookspec
def dbnd_build_project_docker(docker_engine, docker_build_task):
    """Runs on docker build"""



# © Copyright Databand.ai, an IBM Company 2022

"""Module dbnd_run.tasks.basics contains some pipeline examples to start with."""
from dbnd_run.tasks.basics.sanity import dbnd_sanity_check
from dbnd_run.tasks.basics.simplest import SimplestPipeline, SimplestTask


__all__ = [
    "dbnd_sanity_check",
    "bash_cmd",
    "bash_script",
    "SimplestPipeline",
    "SimplestTask",
]
